/*****************************************************************************
 *
 *                        CalC (Calcium Calculator)
 *                             Victor Matveev
 *
 *                              markov.h
 *
 *  Markov variable object (discrete state / continuous time)
 *
 *  Array of markov process objects - class "MarkovArray"
 *  MarkovArray is a member of the "KineticObj" defined in gate.h/gate.cpp
 *
 ******************************************************************************/

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include "syntax.h"
#include "markov.h"

double *MarkovObj::errorTolerance = 0;

//**************************************************************************

double MarkovObj::Evaluate(double Time)
   {
     int    steps, index;
     double Tstep, deltaT, maxProb = 0.0, prob;

     if ( (deltaT = Time - currentTime) <= 0) return Time;
     index = state * (States - 1);

     for (int j=0; j<States; j++) {
       if (j == state) continue;
       if ( (prob = ::Evaluate(transitions[index])) > maxProb )  maxProb = prob;
       index++;
     }

     if (maxProb <= 0.0) return (currentTime = Time);
     
     Tstep = sqrt(*errorTolerance / maxProb);
     steps = ( Tstep >= deltaT ? 1 : int(deltaT / Tstep) );
     if (steps != 1) fprintf(stdlog, "tol=%g maxProb=%g deltaT=%g Tstep=%g steps=%d\n",*errorTolerance,maxProb,deltaT,Tstep,steps);

     for (int subdivisions = 0; subdivisions < steps; subdivisions++) {
       currentTime += Tstep;
       double random = drand48();
       index = state * (States - 1);
       double sum = 0;
       for (int j=0; j<States; j++) {
         if (j == state) continue;
         sum += Tstep * ::Evaluate(transitions[index]);
         // fprintf(stdlog,"transition from %d to %d [%d]: prob=%g cumulative=%g\n",state,j,index,deltaT * ::Evaluate(transitions[index]), sum);
         if (sum > random) { dstate = double(state = j); return Evaluate(currentTime); } 
         index++;
       }
     }
  
     // exit(1);
     // if sum not exceeded, no change   
   }

//*******************************************************************************

MarkovObj::MarkovObj(TokenString &Param, int i)
   {
   long pos = Param.token_index("markov", i+1) - 1;

   ID = Param.getVarName(pos, "markov variable");

   States = Param.get_int(pos+2);
   state  = Param.get_int(pos+3);
   dstate = double(state);
   currentTime = 0.0;

   if (VERBOSE)
       fprintf(stdlog,"  # %s = markov variable with %d states; %s(0)=%d \n", ID, States, ID, state );

   }



//*******************************************************************************

MarkovArray::MarkovArray(TokenString &Param)
{
  long seed = MY_SEED;
  seed =  Param.get_long_param("seed");
  srand48(seed);

  number = Param.token_count("markov");

  array = new MarkovObj *[number];

  if (number && VERBOSE) 
       fprintf(stdlog,"\n### Setting up %d Markov variable(s):\n", number);

  for (int i = 0; i < number; i++)  array[i] = new MarkovObj(Param, i);
}

//*******************************************************************************


void MarkovObj::set_matrix(TokenString &Param, class VarList *VL)
{
  char   *element = 0;
  int    index = 0;
  double value, *ptr;


  transitions = new TermStruct[States*(States-1)];

  for (int i=0; i<States; i++)
    for (int j=0; j<States; j++) {
      if (i == j) continue;
      element = makeMessage("%s.%d.%d",ID,i,j);
      if ( Param.isConst(element, &value) ) {
        transitions[index].type = NUMBER_TYPE;
        transitions[index].val  = value;
      }
      else if (ptr = VL->ResolveID(element) ) {
        transitions[index].type = POINTER_TYPE;
        transitions[index].ptr = ptr;
      }
      else  {   // is zero if undefined
        transitions[index].type = NUMBER_TYPE;
        transitions[index].val  = 0.0;
      }
      delete [] element;       
      index ++;
    }
}


//*******************************************************************************


void MarkovArray::set_matrices(TokenString &Param, class VarList *VL)
{

  for (int i = 0; i < number; i++ )  array[i]->set_matrix(Param, VL);
  
}

//*******************************************************************************
